@@ -270,7 +270,7 @@ namespace strumpack {
   }
 
   template<typename scalar_t> DenseMatrix<scalar_t>
-  DenseMatrix<scalar_t>::transpose() const {
+  DenseMatrix<scalar_t>::conj_transpose() const {
     DenseMatrix<scalar_t> tmp(cols(), rows());
     blas::omatcopy
       ('C', rows(), cols(), data(), ld(), tmp.data(), tmp.ld());
@@ -278,7 +278,7 @@ namespace strumpack {
   }
 
   template<typename scalar_t> void
-  DenseMatrix<scalar_t>::transpose(DenseMatrix<scalar_t>& X) const {
+  DenseMatrix<scalar_t>::conj_transpose(DenseMatrix<scalar_t>& X) const {
     assert(rows() == X.cols() && cols() == X.rows());
     blas::omatcopy
       ('C', rows(), cols(), data(), ld(), X.data(), X.ld());
@@ -750,8 +750,8 @@ namespace strumpack {
     // TODO optimize by implementing by row directly, avoiding transposes
     TIMER_TIME(TaskType::HSS_SEQHQRINTERPOL, 1, t_hss_seq_hqr);
     DenseMatrix<scalar_t> Xt;
-    transpose().ID_column(Xt, piv, ind, rel_tol, abs_tol, max_rank, depth);
-    X = Xt.transpose();
+    conj_transpose().ID_column(Xt, piv, ind, rel_tol, abs_tol, max_rank, depth);
+    X = Xt.conj_transpose();
   }
 
   template<typename scalar_t> void DenseMatrix<scalar_t>::ID_column
@@ -800,13 +800,6 @@ namespace strumpack {
     blas::geqp3tol
       (m, n, tmp.data(), tmp.ld(), ind.data(),
        tau.get(), rank, rel_tol, abs_tol);
-    std::vector<int> piv(n);
-    for (int i=1; i<=n; i++) {
-      int j = ind[i-1];
-      assert(j-1 >= 0 && j-1 < int(ind.size()));
-      while (j < i) j = ind[j-1];
-      piv[i-1] = j;
-    }
     V = DenseMatrix<scalar_t>(rank, cols(), tmp.ptr(0, 0), tmp.ld());
     for (int c=0; c<rank; c++)
       for (int r=c+1; r<rank; r++)
